import skil_client
import time
import uuid
import numpy as np

class Service:
    '''A wrapper around a deployed model for inference.
    '''
    def __init__(self, skil, model_name, deployment, model_deployment):
        self.skil = skil
        self.model_name = model_name
        self.model_deployment = model_deployment
        self.deployment = deployment

    def start(self):
        '''Starts the service.
        '''
        if not self.model_deployment:
            self.skil.printer.pprint(
                "No model deployed yet, call 'deploy()' on a model first.")
        else:
            self.skil.api.model_state_change(
                self.deployment.id,
                self.model_deployment.id,
                skil_client.SetState("start")
            )

            self.skil.printer.pprint(">>> Starting to serve model...")
            while True:
                time.sleep(5)
                model_state = self.skil.api.model_state_change(
                    self.deployment.id,
                    self.model_deployment.id,
                    skil_client.SetState("start")
                ).state
                if model_state == "started":
                    time.sleep(15)
                    self.skil.printer.pprint(
                        ">>> Model server started successfully!")
                    break
                else:
                    self.skil.printer.pprint(">>> Waiting for deployment...")


    def stop(self):
        '''Stop the service.
        '''
        # TODO: test this
        self.skil.api.model_state_change(
            self.deployment.id,
            self.model_deployment.id,
            skil_client.SetState("stop")
        )

    def _indarray(self, np_array):
        '''Convert a numpy array to `skil_client.INDArray` instance.

        # Arguments
        np_array: `numpy.ndarray` instance.

        # Returns
        `skil_client.INDArray` instance.
        '''
        return skil_client.INDArray(
            ordering='c',
            shape=list(np_array.shape),
            data = np_array.reshape(-1).tolist()
        )

    def predict(self, data):
        '''Predict for given batch of data.

        # Argments
        data: `numpy.ndarray` (or list thereof). Batch of input data, or list of batches for multi-input model.

        # Returns
        `numpy.ndarray` instance for single output model and list of `numpy.ndarray` for multi-ouput model.
        '''
        if isinstance(data, list):
            inputs = [self._indarray(x) for x in data]
        else:
            inputs = [self._indarray(data)]

        classification_response = self.skil.api.multipredict(
            deployment_name=self.deployment.name,
            model_name=self.model_name,
            version_name="default",
            body=skil_client.MultiPredictRequest(
                id=str(uuid.uuid1()),
                needs_pre_processing=False,
                inputs=inputs
            )
        )
        outputs = classification_response.outputs
        outputs = [np.asarray(o.data).reshape(o.shape) for o in outputs]
        if len(outputs) == 1:
            return outputs[0]
        return outputs

    def predict_single(self, data):
        '''Predict for a single input.

        # Argments
        data: `numpy.ndarray` (or list thereof). Input data.

        # Returns
        `numpy.ndarray` instance for single output model and list of `numpy.ndarray` for multi-ouput model.
        '''
        if isinstance(data, list):
            inputs = [self._indarray(np.expand_dims(x, 0)) for x in data]
        else:
            inputs = [self._indarray(np.expand_dims(data, 0))]

        classification_response = self.skil.api.multipredict(
            deployment_name=self.deployment.name,
            model_name=self.model_name,
            version_name="default",
            body=skil_client.MultiPredictRequest(
                id=str(uuid.uuid1()),
                needs_pre_processing=False,
                inputs=inputs
            )
        )
        output = classification_response.outputs[0]
        return np.asarray(output.data).reshape(output.shape)



import skil_client


class Deployment:
    """ Deployments operate independently of workspaces to ensure that there are no accidental interruptions or mistakes in a production environment.

    # Arguments:
    skil: Skil server instance.
    name: string. Name for the deployment.
    id: Unique id for the deployment. If `None`, a unique id will be generated.
    """
    def __init__(self, skil, name=None, id=None):
        if id is not None:
            response = skil.api.deployment_get(id)
            if response is None:
                raise KeyError('Deployment not found: ' + str(id))
            self.response = response
            self.name = self.response.name
        else:
            self.name = name if name else 'deployment'
            create_deployment_request = skil_client.CreateDeploymentRequest(self.name)
            self.response = skil.api.deployment_create(create_deployment_request)
            self.id = self.response.id


def get_deployement_by_id(skil, id):
    dep = Deployment(skil, id=id)
    return dep



from .experiments import Experiment
import skil_client


class WorkSpace:
    """Workspaces are a collection of features that enable different tasks such as conducting experiments, training models, and test different dataset transforms.

    Workspaces are distinct from Deployments by operating as a space for non-production work.

    # Arguments
    skil: Skil server instance
    name: string. Name for the workspace.
    labels: string. Labels associated with the workspace, useful for searching (comma seperated).
    verbose: boolean. If True, api response will be printed.
    create: boolean. Internal, do not use.
        
    """
    def __init__(self, skil=None, name=None, labels=None, verbose=False, create=True):
        if not create:
            return
        self.skil = skil
        self.printer = self.skil.printer
        self.name = name if name else 'skil_workspace'

        self.workspace = self.skil.api.add_model_history(
            self.skil.server_id,
            skil_client.AddModelHistoryRequest(name, labels)
        )
        self.id = self.workspace.model_history_id

        if verbose:
            self.printer.pprint(self.workspace)

    def delete(self):
        """Deletes the work space.
        """
        try:
            api_response = self.skil.api.delete_model_history(
                self.skil.server_id, self.workspace.id)
            self.skil.printer.pprint(api_response)
        except skil_client.rest.ApiException as e:
            self.skil.printer.pprint(
                ">>> Exception when calling delete_model_history: %s\n" % e)

def get_workspace_by_id(self, skil, id):
    server_id = skil.server_id
    response = skil.api.get_model_history(server_id, id)
    ws = WorkSpace(create=False)
    ws.skil = skil
    ws.printer = skil.printer
    ws.workspace = response
    ws.id = id
    ws.name = response.model_name
    return ws

def add_experiment(self, id=None, name='test', description='test', verbose=False):
    return Experiment(self, id=id, name=name, description=description, verbose=verbose)



import skil
from skil.services import Service

import skil_client
import time
import os
import uuid


class Model:
    """
    SKIL wrapper for DL4j, Keras and TF models

    SKIL has a robust model storage, serving, and import system for supporting major deep learning libraries.
    SKIL can be used for end-to-end training, configuration, and deployment of models or alternatively you can import models into SKIL.

    # Arguments
    model: string. Model file path.
    id: integer. Unique id for model. If `None`, a unique id will be generated.
    name: string. Name for the model.
    version: integer. Version of the model. Defaults to 1.
    experiment: `Experiment` instance. If `None`, an `Experiment` object will be created internally.
    labels: string. Labels associated with the workspace, useful for searching (comma seperated).
    verbose: boolean. If `True`, prints api response.
    create: boolean. Internal. Do not use.
    """
    def __init__(self, model=None, id=None, name=None, version=None, experiment=None,
                 labels='', verbose=False, create=True):
        if create:
            if isinstance(model, str) and os.path.isfile(model):
                model_file_name = model
            else:
                if hasattr(model, 'save'):
                    model_file_name = 'temp_model.h5'
                    if os.path.isfile(model_file_name):
                        os.remove(model_file_name)
                    model.save(model_file_name)
                else:
                    raise Exception('Invalid model: ' + str(model))
            if not experiment:
                self.skil = skil.Skil()
                self.work_space = skil.workspaces.WorkSpace(self.skil)
                self.experiment = skil.experiments.Experiment(self.work_space)
            else:
                self.experiment = experiment
                self.work_space = experiment.work_space
                self.skil = self.work_space.skil
            self.skil.upload_model(os.path.join(os.getcwd(), model_file_name))

            self.model_name = model_file_name
            self.model_path = self.skil.get_model_path(model_file_name)
            self.id = id if id else uuid.uuid1()
            self.name = name if name else model_file_name
            self.version = version if version else 1

            self.evaluations = {}

            add_model_instance_response = self.skil.api.add_model_instance(
                self.skil.server_id,
                skil_client.ModelInstanceEntity(
                    uri=self.model_path,
                    model_id=id,
                    model_labels=labels,
                    model_name=name,
                    model_version=self.version,
                    created=int(round(time.time() * 1000)),
                    experiment_id=self.experiment.id
                )
            )
            if verbose:
                self.skil.printer.pprint(add_model_instance_response)
        else:
            self.experiment = experiment
            self.work_space = experiment.work_space
            self.skil = self.work_space.skil
            assert id is not None
            self.id = id
            model_entity = self.skil.api.get_model_instance(self.skil.server_id,
                                                            id)
            self.name = model_entity.model_name
            self.version = model_entity.model_version
            self.model_path = self.skil.get_model_path(model_file_name)
            self.model = None

    def delete(self):
        """Deletes the model
        """
        try:
            self.skil.api.delete_model_instance(self.skil.server_id, self.id)
        except skil_client.rest.ApiException as e:
            self.skil.printer.pprint(
                ">>> Exception when calling delete_model_instance: %s\n" % e)

    def add_evaluation(self, accuracy, id=None, name=None, version=None):
        eval_version = version if version else 1
        eval_id = id if id else self.id
        eval_name = name if name else self.id

        eval_response = self.skil.api.add_evaluation_result(
            self.skil.server_id,
            skil_client.EvaluationResultsEntity(
                evaluation="",  # TODO: what is this?
                created=int(round(time.time() * 1000)),
                eval_name=eval_name,
                model_instance_id=self.id,
                accuracy=float(accuracy),
                eval_id=eval_id,
                eval_version=eval_version
            )
        )
        self.evaluations[id] = eval_response

    def deploy(self, deployment=None, start_server=True, scale=1, input_names=None,
               output_names=None, verbose=True):
        """Deploys the model

        # Arguments:
        deployment: `Deployment` instance.
        start_server: boolean. If `True`, the service is immedietely started.
        scale: integer. Scale for deployment.
        input_names: list of strings. Input variable names of the model.
        output_names: list of strings. Output variable names of the model.
        verbose: boolean. If `True`, api response will be printed.

        # Returns:
        `Service` instance.
        """
        if not deployment:
            deployment = skil.Deployment(skil=self.skil, name=self.name)

        uris = ["{}/model/{}/default".format(deployment.name, self.name),
                "{}/model/{}/v1".format(deployment.name, self.name)]

        deploy_model_request = skil_client.ImportModelRequest(
            name=self.name,
            scale=scale,
            file_location=self.model_path,
            model_type="model",
            uri=uris,
            input_names=input_names,
            output_names=output_names)

        self.deployment = deployment.response
        self.model_deployment = self.skil.api.deploy_model(
            self.deployment.id, deploy_model_request)
        if verbose:
            self.skil.printer.pprint(self.model_deployment)

        service = Service(self.skil, self.name, self.deployment, self.model_deployment)
        if start_server:
            service.start()
        return service

    def undeploy(self):
        """Undeploy the model.
        """
        try:
            self.skil.api.delete_model(self.deployment.id, self.id)
        except skil_client.rest.ApiException as e:
            self.skil.printer.pprint(
                ">>> Exception when calling delete_model_instance: %s\n" % e)


def get_model_by_id(self, experiment, id):
    return Model(id=id, experiment=experiment, create=False)



import skil_client


class Experiment:
    """Experiments in SKIL are useful for defining different model configurations, encapsulating training of models, and carrying out different data cleaning tasks.

    Experiments have a one-to-one relationship with Notebooks and have their own storage mechanism for saving different model configurations when seeking a best candidate.

    # Arguments:
    work_space: `WorkSpace` instance.
    id: integer. Unique id for workspace. If `None`, a unique id will be generated.
    name: string. Name for the experiment.
    description: string. Description for the experiment.
    verbose: boolean. If `True`, api response will be printed.
    """
    def __init__(self, work_space=None, id=None, name='test', description='test', verbose=False, create=True):
        if create:
            self.work_space = work_space
            self.skil = self.work_space.skil
            self.id = id if id else work_space.id + "_experiment"
            self.name = name
            experiment_entity = skil_client.ExperimentEntity(
                                            experiment_id=self.id,
                                            experiment_name=name,
                                            experiment_description=description,
                                            model_history_id=self.work_space.id
                                            )
    
            add_experiment_response = self.skil.api.add_experiment(
                self.skil.server_id,
                experiment_entity
            )

            self.experiment_entity = experiment_entity

            if verbose:
                self.skil.printer.pprint(add_experiment_response)
        else:
            experiment_entity = work_space.skil.api.get_experiment(
                work_space.skil.server_id,
                id
            )
            self.experiment_entity = experiment_entity
            self.work_space = work_space
            self.id = id
            self.name = experiment_entity.experiment_name

    def delete(self):
        """Deletes the experiment.
        """
        try:
            api_response = self.skil.api.delete_experiment(
                self.work_space.id, self.id)
            self.skil.printer.pprint(api_response)
        except skil_client.rest.ApiException as e:
            self.skil.printer.pprint(
                ">>> Exception when calling delete_experiment: %s\n" % e)

def get_experiment_by_id(work_space, id):
    return Experiment(work_space=work_space, id=id, create=False)



import os
import shutil
import sys
import uuid

import keras

try:
  basestring
except NameError:
  basestring = str


class SkilContext(object):
    '''SkilContext manages models from within a Zeppelin notebook
    hosted by SKIL. The Spark context necessary for initialization
    is provided in any notebook created with SKIL.

    SkilContext can upload models, add them to an experiment and
    add evaluation metrics to a model.  
    '''
    def __init__(self, sc):
        self._sc = sc

        self.JvmSkilContext = sc._jvm.io.skymind.zeppelin.utils.SkilContext
        self._ctx = self.JvmSkilContext()

        self.SKILEnvironment = sc._jvm.io.skymind.skil.service.SKILEnvironment
        self.ModelInstanceEntity = sc._jvm.io.skymind.modelproviders.history.model.ModelInstanceEntity
        self.Nd4j = sc._jvm.org.nd4j.linalg.factory.Nd4j
        self.Evaluation = sc._jvm.org.deeplearning4j.eval.Evaluation


    def _models_path(self):
        service_path = self.SKILEnvironment.skilServiceWorkingDirFile().toString()
        storage_path = os.path.join(service_path, "storage")
        models_path = os.path.join(storage_path, "models")

        if not os.path.exists(storage_path):
            os.mkdir(storage_path)
        if not os.path.exists(models_path):
            os.mkdir(models_path)

        return models_path

    def experiment_id(self, z):
        '''Get the experiment ID of the current notebook or None.
        :param z: the ZeppelinContext
        :return: the experiment ID
        '''

        return self._ctx.experiment_id(z.z)

    def save_model(self, z, model):
        '''Save the model into the managed models directory.
        :param z: The ZeppelinContext
        :param model: The model to save
        :return: The path of the saved model.
        '''

        experiment_id = self.experiment_id(z)
        if experiment_id is None:
            return None

        if isinstance(model, keras.models.Model):
            models_path = self._models_path()
            file_path = os.path.join(models_path, str(uuid.uuid1()) + '.h5')
            
            if isinstance(file_path, unicode):
                file_path = file_path.encode(sys.getfilesystemencoding())

            model.save(file_path)

            return file_path
        else:
            raise NotImplementedError("Only Keras models currently supported.")

    def copy_model(self, z, source_path, model_type):
        '''Copy a model file (tensorflow or ONNX) to the managed model directory.
        :param z: The ZeppelinContext
        :param path: The path to the model you want to copy
        :param model_type: The type of model. Currently either 'tensorflow' or 'onnx'
        :return: The path of the saved model
        '''

        models_path = self._models_path()

        if model_type.lower() == 'tensorflow' or model_type.lower() == 'tf':
            dest_path = os.path.join(models_path, str(uuid.uuid1()) + '.pb')
        elif model_type.lower() == 'onnx':
            dest_path = os.path.join(models_path, str(uuid.uuid1()) + '.onnx')
        else:
            raise NotImplementedError('Only TensorFlow and ONNX model types are supported.')

        shutil.copyfile(source_path, dest_path)

        return dest_path

    def add_model_to_experiment(self, z, model, name=None):
        '''
        Add the model to the model list inside the SKIL Experiment.
        :param z: The ZeppelinContext
        :param model: The model to save or path.
        :param name: Optional name of the model.
        :return: The ModelInstanceID for use with adding EvaluationResults.
        '''

        model_id = str(uuid.uuid1())
        model_path = None
        if isinstance(model, basestring):
            model_path = model
        elif isinstance(model, keras.models.Model):
            model_path = self.save_model(z, model)
        else:
            raise NotImplementedError("Can only auto-save Keras models. For TensorFlow or ONNX models please save them separately.")

        instance = (self.ModelInstanceEntity
            .builder()
            .modelName(name)
            .experiment_id(self.experiment_id(z))
            .modelId(model_id)
            .created(self._sc._jvm.java.util.Date())
            .uri('file://' + model_path)
            .etlJson(None)
            .notebookJson(None)
            .build())

        self._ctx.getClient().addModelInstance(instance)

        return model_id

    def add_evaluation_to_model(self, z, model_id, model, data, labels, name=None):
        '''
        Gathers the evalutation results of the model on the specified test data.
        :param z: The ZeppelinContext
        :param model_id: The ModelInstanceID
        :param model: The model to use
        :param data: The dataset to evaluate
        :param labels: The labels of the dataset
        :param name: A name for the EvaluationResults
        :return: The id of the EvaluationResults if saved.
        '''

        def assign(np_arr, jvm_arr):
            i = 0
            for row in np_arr.tolist():
                j = 0
                for col in row:
                    jvm_arr[i][j] = col
                    j += 1
                i += 1

        if not isinstance(model, keras.models.Model):
            raise NotImplementedError("Only Keras models currently supported.")

        preds = model.predict(data)
        p_arr = self._sc._gateway.new_array(self._sc._gateway.jvm.double, preds.shape[0], preds.shape[1])
        assign(preds, p_arr)
        nd_pred = self.Nd4j.create(p_arr)
        y_arr = self._sc._gateway.new_array(self._sc._gateway.jvm.double, labels.shape[0], labels.shape[1])
        assign(labels, y_arr)
        nd_y = self.Nd4j.create(y_arr)

        e = self.Evaluation(int(nd_y.shape()[1]))
        e.eval(nd_y, nd_pred)

        return self._ctx.addEvaluationToModel(z.z, model_id, e, name)

    # Camel case methods for backward compatibility
    def experimentId(self, z):
        return self.experiment_id(z)
    
    def saveModel(self, z, model):
        return self.save_model(z, model)

    def copyModel(self, z, source_path, model_type):
        return self.copy_model(z, source_path, model_type)

    def addModelToExperiment(self, z, model, name=None):
        return self.add_model_to_experiment(z, model, name)

    def addEvaluationToModel(self, z, model_id, model, data, labels, name=None):
        return self.add_evaluation_to_model(z, model_id, model, data, labels, name)



from setuptools import setup
from setuptools import find_packages

setup(
    name='skil',
    version='0.2.5',
    packages=find_packages(),
    install_requires=['skil_client', 'requests', 'numpy'],
    extras_require={
        'tests': ['pytest', 'pytest-pep8', 'pytest-cov', 'mock']
    },
    include_package_data=True,
    license='Apache',
    description='Deploy your Python models with SKIL',
    long_description='Deploy your Python models with SKIL',
    author='Max Pumperla',
    author_email='max@skymind.io',
    url='https://github.com/SkymindIO/skil-python',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ]
)



from .workspaces import WorkSpace
import skil_client
import pprint
import os
import time
import requests
import json
import subprocess


def start_skil_docker():
    devnull = open(os.devnull, 'w')

    print(">>> Downloading latest SKIL docker image.")
    subprocess.call(["docker", "pull", "skymindops/skil-ce"])
    print(">>> Starting SKIL docker container.")
    subprocess.Popen(["docker", "run", "--rm", "-it", "-p", "9008:9008",
                      "-p", "8080:8080", "skymindops/skil-ce", "bash", "/start-skil.sh", "&"],
                     stdout=devnull, stderr=subprocess.STDOUT)
    print("Starting SKIL. This process will take a few seconds to start.")
    time.sleep(20)
    print("SKIL started! Visit http://localhost:9008 to work with the UI.")


class Skil:
    def __init__(self, workspace_server_id=None, host='localhost', port=9008,
                 debug=False, user_id='admin', password='admin'):

        self.printer = pprint.PrettyPrinter(indent=4)

        config = skil_client.Configuration()
        config.host = "{}:{}".format(host, port)
        config.debug = debug
        self.config = config
        self.uploads = []
        self.uploaded_model_names = []
        self.auth_headers = None

        self.api_client = skil_client.ApiClient(configuration=config)
        self.api = skil_client.DefaultApi(api_client=self.api_client)

        try:
            self.printer.pprint('>>> Authenticating SKIL...')
            credentials = skil_client.Credentials(
                user_id=user_id, password=password)
            token = self.api.login(credentials)
            self.token = token.token
            config.api_key['authorization'] = self.token
            config.api_key_prefix['authorization'] = "Bearer"
            self.printer.pprint('>>> Done!')
        except skil_client.rest.ApiException as e:
            raise Exception(
                "Exception when calling DefaultApi->login: {}\n".format(e))

        if workspace_server_id:
            self.server_id = workspace_server_id
        else:
            self.server_id = self.get_default_server_id()

    def get_default_server_id(self):
        self.auth_headers = {'Authorization': 'Bearer %s' % self.token}
        r = requests.get(
            'http://{}/services'.format(self.config.host), headers=self.auth_headers)
        if r.status_code != 200:
            r.raise_for_status()

        content = json.loads(r.content.decode('utf-8'))
        services = content.get('serviceInfoList')
        for s in services:
            if 'Model History' in s.get('name'):
                id = s.get('id')
        if id:
            return id
        else:
            raise Exception(
                "Could not detect default model history server instance. Is SKIL running?")

    def upload_model(self, model_name):
        self.printer.pprint('>>> Uploading model, this might take a while...')
        upload = self.api.upload(file=model_name).file_upload_response_list
        self.uploads = self.uploads + upload
        self.uploaded_model_names.append(model_name)
        self.printer.pprint(self.uploads)

    def get_uploaded_model_names(self):
        return self.uploaded_model_names

    def get_model_path(self, model_name):
        for upload in self.uploads:
            if model_name == upload.file_name:
                return "file://" + upload.path
        raise Exception("Model resource not found, did you upload it? ")

    def add_work_space(self, name=None, labels=None, verbose=False):
        return WorkSpace(self, name=name, labels=labels, verbose=verbose)
